
// Oru class
#import "DirectAction.h"

// Project classes
#import "XMLComponent.h"
#import "NSScanner-FarmExtensions.h"
#import <Foundation/NSDebug.h>


@implementation DirectAction : WODirectAction 

static NSString *		_cachedConfigResult = nil;
static NSString *		_cachedConfigPath = nil;
static NSDate *			_cachedConfigModificationDate = nil;


- (WOComponent *)defaultAction
{
    id	nextPage = [self pageWithName:@"Main"];
    [nextPage takeValue:[self xmlRepresentationOfConfig] forKey:@"hostConfig"];
    
    return nextPage;
}

- (WOComponent *)woconfigAction
{
    id		nextPage = [self pageWithName:@"XMLComponent"];

    [nextPage setXMLContent:[self xmlRepresentationOfConfig]];

    return nextPage;
}

- (NSString *)xmlRepresentationOfConfig
/*"
    Converts the oldConfig string into a XML string according to the 4.5 WebObjects.xml specification.
    Missing configuration parameters are retrieved from the user defaults.
"*/
{
    NSString * 		resultString;

    // First we need to build a dictionary from the files contents.

    id		configPath = [[NSUserDefaults standardUserDefaults] stringForKey:@"WebObjectsSourceConfig"];
    id		configContents;
    id		fileInfo = [[NSFileManager defaultManager] fileAttributesAtPath:configPath traverseLink:YES];

    // Lets see if we can reuse the cache.

    if( [_cachedConfigPath isEqual:configPath] &&
        [_cachedConfigModificationDate isEqual:[fileInfo objectForKey:NSFileModificationDate]]  )
    {
        return _cachedConfigResult;
    }
    // Well .. caching failed ... so lets read it and remember what we read.
    // this is critical ...s o lets lock the request handling first...

    // [_cacheUpdateLock lockRequestHandling];

    [_cachedConfigPath release];
    _cachedConfigPath = [configPath retain];

    [_cachedConfigModificationDate release];
    _cachedConfigModificationDate = [[fileInfo objectForKey:NSFileModificationDate] retain];

    [_cachedConfigResult release];

    configContents = [self _parseConfigFile:configPath];

    // Now first build 
    resultString = @"<?xml version=\"1.0\" encoding=\"ASCII\"?>\n";

    resultString = [resultString stringByAppendingString:[self _xmlAdaptorConfiguration]];
    resultString = [resultString stringByAppendingString:[self _xmlApplicationConfiguration:configContents]];
    resultString = [resultString stringByAppendingString:@"</adaptor>\n"];

    _cachedConfigResult = [resultString retain];

    if( NSDebugEnabled ) NSLog( @"Parsed %@ and produced this XML config:\n %@", _cachedConfigPath,_cachedConfigResult );

    // [_cacheUpdateLock unlockRequestHandling];

    return _cachedConfigResult;
}

- (NSDictionary *)_parseConfigFile:(NSString *)configPath
/*"
    Returns a dictionary containing the parsed content of the old-style WebObjects.conf file
    at the provided configPath.
"*/
{
    NSString *			oldConfig = [NSString stringWithContentsOfFile:configPath];
    NSScanner *			lineScanner = [NSScanner scannerWithString:oldConfig];
    NSCharacterSet *	lineEndChars = [NSCharacterSet characterSetWithCharactersInString:@"\n\r"];
    NSString *			nextLine;

    id				configContents = [NSMutableDictionary new];

    // Now walk over the conf string and pick out the single lines of interest.

    while( [lineScanner isAtEnd] == NO )
    {
        id 	lineContent;
        id	appName;
        id	appDict = nil;
        id	instanceDict;

        // Remove leading blanks...
        
        [lineScanner scanCharactersFromSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]
                                intoString:NULL];
        if( [lineScanner isAtEnd] ) break;

        // Now get the rest
        
        [lineScanner scanUpToCharactersFromSet:lineEndChars intoString:&nextLine skipSet:YES];

        lineContent = [self _parseConfigLine:nextLine];

        appName = [lineContent objectForKey:@"name"];
        appDict = [configContents objectForKey:appName];

        // if this app is still unknown lets prepare its entry
        
        if( appDict == nil )
        {
            appDict = [NSMutableDictionary dictionary];
            [configContents setObject:appDict forKey:appName];
            [appDict setObject:[NSMutableArray array] forKey:@"instances"];
        }

        // Now build a dictionary for the instance...

        instanceDict = [NSMutableDictionary dictionary];
        [instanceDict setObject:[lineContent objectForKey:@"id"] forKey:@"id"];
        [instanceDict setObject:[lineContent objectForKey:@"host"] forKey:@"host"];
        [instanceDict setObject:[lineContent objectForKey:@"port"] forKey:@"port"];

        [[appDict objectForKey:@"instances"] addObject:instanceDict];
    }
    return [configContents autorelease];
}

- (NSDictionary *)_parseConfigLine:(NSString *)anEntry
/*"
 Parses a line of the format   name:id@host port [{ key=value; ...}]   and returns the contents in
 form of a dictionry. The provided anEntry string must be "clean" at the start..on white space please.
 The line to scan looks like this:  MyApp:1@woa 2001 
"*/
{
    NSMutableDictionary *	content = [NSMutableDictionary new];
    NSString *		someString;
    NSScanner *		ourScanner = [NSScanner scannerWithString:anEntry];

    // Get the instance NAME...

    [ourScanner scanUpToCharactersFromSet:[NSCharacterSet characterSetWithCharactersInString:@": \t"]
                               intoString:&someString
                                  skipSet:YES];
    [content setObject:someString forKey:@"name"];

    // Get the instance ID...

    [ourScanner scanUpToCharactersFromSet:[NSCharacterSet characterSetWithCharactersInString:@"@ \t"]
                               intoString:&someString
                                  skipSet:YES];
    [content setObject:someString forKey:@"id"];

    // Get the instance HOST...

    [ourScanner scanUpToCharactersFromSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]
                               intoString:&someString
                                  skipSet:YES];
    [content setObject:someString forKey:@"host"];

    // Get the instance PORT...

    [ourScanner scanUpToCharactersFromSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]
                               intoString:&someString
                                  skipSet:YES];
    [content setObject:someString forKey:@"port"];

    return [content autorelease];
}

- (NSString *)_xmlAdaptorConfiguration
/*"
  Returns a string with the adaptor part of the XML configuration based on the values from the user defaults.
 "*/
{
    id		result = [NSMutableString stringWithString: @"<adaptor"];

    // Now check all known adaptor attributes ... but we will not validate the values.
    
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"retries"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"scheduler"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"dormant"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"protocol"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"transport"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"redir"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"xyzzy"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"confinterval"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"timeout"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"sendTimeout"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"recvTimeout"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"cnctTimeout"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"poolSize"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"sendBufSize"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"recvBufSize"];
    [self _appendToXMLAdaptorConfiguration:result defaultForKey:@"urlVersion"];

    [result appendString:@">\n"];

    return result;
}

- (void)_appendToXMLAdaptorConfiguration:(NSMutableString *)mutableContent defaultForKey:(NSString *)aKey
{
    id		userDefaults = [NSUserDefaults standardUserDefaults];
    id		aValue = [userDefaults stringForKey:[@"WebObjects." stringByAppendingString:aKey]];

    if( aValue )
    {
        // <<HACK>> Sorry ... this is bound to fail as soon as we get some spaces in any of the entries.

        [mutableContent appendFormat:@" %@ =\"%@\"", aKey, aValue];
    }
}

- (NSString *)_xmlApplicationConfiguration:(NSDictionary *)configurationDict
{
    id		result = @"";
    id		nextService = nil;
    id		enumerator = [configurationDict keyEnumerator];
   
    while( nextService = [enumerator nextObject] )
    {
        id		serviceDict = [configurationDict objectForKey:nextService];
        result = [result stringByAppendingString:[self _xmlForService:nextService
                                                           withConfig:serviceDict]];
    }

    return result;
}

- (NSString *)_xmlForService:(NSString *)name withConfig:(NSDictionary *)appDict
/*"
Returns of string with the nature of the following XML structure:
     
    <application scheduler="roundrobin" name="FarmersWebMail">
        <instance port="2001" host="panda" id="1"/>
        <instance port="2004" host="panda" id="4"/>
    </application>
"*/

{
    id		instances = [appDict objectForKey:@"instances"];
    id		instanceEnum = [instances objectEnumerator];
    id		hostMappingDict = [[NSUserDefaults standardUserDefaults] dictionaryForKey:@"HostMapping"];
    id		anInstance;
    id		result = [@"   <application" mutableCopy];

    // Maybe there is nothing to do ... well it would be strange since that would mean we have
    // a potentially corrupted config file.
    
    if( [instances count] == 0 ) return @"";

    // First put in the global app information
    
    [result appendFormat:@" name=\"%@\"", name];

    [self _appendToXMLApplicationConfiguration:result defaultForKey:@"retries" forService:name];
    [self _appendToXMLApplicationConfiguration:result defaultForKey:@"scheduler" forService:name];
    [self _appendToXMLApplicationConfiguration:result defaultForKey:@"dormant" forService:name];
    [self _appendToXMLApplicationConfiguration:result defaultForKey:@"protocol" forService:name];
    [self _appendToXMLApplicationConfiguration:result defaultForKey:@"transport" forService:name];
    [self _appendToXMLApplicationConfiguration:result defaultForKey:@"redir" forService:name];
    [self _appendToXMLApplicationConfiguration:result defaultForKey:@"timeout" forService:name];
    [self _appendToXMLApplicationConfiguration:result defaultForKey:@"sendTimeout" forService:name];
    [self _appendToXMLApplicationConfiguration:result defaultForKey:@"recvTimeout" forService:name];
    [self _appendToXMLApplicationConfiguration:result defaultForKey:@"cnctTimeout" forService:name];
    [self _appendToXMLApplicationConfiguration:result defaultForKey:@"poolSize" forService:name];
    [self _appendToXMLApplicationConfiguration:result defaultForKey:@"urlVersion" forService:name];

    [result appendString:@">\n"];

    // Now add the instance information

    while( anInstance = [instanceEnum nextObject] )
    {
        id		instanceName = [NSString stringWithFormat:@"%@.%@", name, [anInstance objectForKey:@"port"]];
        id		hostName = [anInstance objectForKey:@"host"];
        id		mappedHostName = [hostMappingDict objectForKey:hostName];
        
        [result appendString:@"      <instance"];

        [result appendFormat:@" port=\"%@\"", [anInstance objectForKey:@"port"]];
        if( mappedHostName != nil ) hostName = mappedHostName;
        [result appendFormat:@" host=\"%@\"", hostName];
        [result appendFormat:@" id=\"%@\"", [anInstance objectForKey:@"id"]];

        // Now append potential defaults for this instance...

        [self _appendToXMLApplicationConfiguration:result defaultForKey:@"dormant" forService:instanceName];
        [self _appendToXMLApplicationConfiguration:result defaultForKey:@"protocol" forService:instanceName];
        [self _appendToXMLApplicationConfiguration:result defaultForKey:@"transport" forService:instanceName];
        [self _appendToXMLApplicationConfiguration:result defaultForKey:@"redir" forService:instanceName];
        [self _appendToXMLApplicationConfiguration:result defaultForKey:@"timeout" forService:instanceName];
        [self _appendToXMLApplicationConfiguration:result defaultForKey:@"sendTimeout" forService:instanceName];
        [self _appendToXMLApplicationConfiguration:result defaultForKey:@"recvTimeout" forService:instanceName];
        [self _appendToXMLApplicationConfiguration:result defaultForKey:@"cnctTimeout" forService:instanceName];
        [self _appendToXMLApplicationConfiguration:result defaultForKey:@"poolSize" forService:instanceName];
        [self _appendToXMLApplicationConfiguration:result defaultForKey:@"urlVersion" forService:instanceName];
        [self _appendToXMLApplicationConfiguration:result defaultForKey:@"refuseNewSessions" forService:instanceName];

        [result appendString:@"/>\n"];
    }
    [result appendString:@"   </application>\n"];

    return [result autorelease];
}

- (void)_appendToXMLApplicationConfiguration:(NSMutableString *)mutableContent
                               defaultForKey:(NSString *)aKey
                                  forService:(NSString *)name;
{
    id		userDefaults = [NSUserDefaults standardUserDefaults];
    id		aValue = [userDefaults stringForKey:[NSString stringWithFormat:@"WebObjects.%@.%@", name, aKey]];

    if( aValue )
    {
        // <<HACK>> Sorry ... this is bound to fail as soon as we get some spaces in any of the entries.

        [mutableContent appendFormat:@" %@=\"%@\"", aKey, aValue];
    }
}


@end
